#include <stdio.h>
#include <stdarg.h>

#include "defs/defs.h"	//ifmod
#include "lib/bvhfile.h"

//#define USE_ISOSURF
//#define USE_ISOSURF_REALTIME

#ifdef USE_ISOSURF
#include "lib/isurface.h"
#endif

//rings per bone
#define BONE_DETAIL			3
//vertices per ring
#define VERTICES_PER_BONE	5

BVHNode::vtx_ring_t BVHNode::genVertexRing(
	std::vector<BVHNode::vtx_t> &vtxList,
	int node,
	float coeff,
	const vec3f &normal)
{
	vec3f tangent = vec3unit(cross(normal, vec3f(0,0,1)));
	vec3f binormal = vec3unit(cross(normal, tangent));

	vtx_ring_t ret;
	ret.first = (int)vtxList.size();
	ret.count = VERTICES_PER_BONE;
	vtxList.resize(ret.first + ret.count);

	//adjust params to reshape curve
	float bone_radius = 1.f + exp(-(coeff-0.5f)*(coeff-0.5f) * 2.f) * 2.f;

	if (getType() == BVH_NODETYPE_ENDSITE) {
		bone_radius *= pow(1.f-coeff, 0.1f);
	}

	for (int i = 0; i < ret.count; i++) {
		float theta = 2.f * (float)M_PI * (float)i / (float)ret.count;
		float costh = (float)cos(theta) * bone_radius;
		float sinth = (float)sin(theta) * bone_radius;
		vtx_t &v = vtxList[ret.first + i];
		v.x = tangent * costh + binormal * sinth;
		v.n = node;
		v.t = coeff;
	}

	return ret;
}


void BVHNode::bindVertexRing(
	std::vector<int> &triList,
	const BVHNode::vtx_ring_t &r1,
	const BVHNode::vtx_ring_t &r2)
{
	//write poly between the vertices

	int i;
	for (i = 0; i < r1.count; i++) {
		int i2 = i;
		if (i2 >= r2.count) i2 = r2.count - 1;
		triList.push_back(r1.first + i);
		triList.push_back(r2.first + i2);
		triList.push_back(r1.first + ((i+1) % r1.count));
	}

	for (i = 0; i < r2.count; i++) {
		int i2 = i+1;
		if (i2 > r1.count) i2 = r1.count;
		if (i2 == r1.count) i2 = 0;
		triList.push_back(r1.first + i2);
		triList.push_back(r2.first + i);
		triList.push_back(r2.first + ((i+1) % r2.count));
	}
}

BVHNode::vtx_ring_t BVHNode::genRing(
	std::vector<BVHNode *> &nodeList,
	std::vector<int> &triList,
	std::vector<BVHNode::vtx_t> &vtxList)
{
	vtx_ring_t firstRing, dstRing;

	if (parentIndex > -1) {

		BVHNode *p = nodeList[parentIndex];

		vec3f normal = vec3unit(p->basis.t - basis.t);

		firstRing = dstRing = genVertexRing(vtxList, nodeIndex, 1.f / BONE_DETAIL, normal);

		for (int i = 0; i < BONE_DETAIL; i++) {
			vtx_ring_t srcRing = dstRing;
			dstRing = genVertexRing(vtxList, nodeIndex, (float)(i + 1) / (float)BONE_DETAIL, normal);
			bindVertexRing(triList, srcRing, dstRing);
		}
	} else {
		firstRing = dstRing = genVertexRing(vtxList, nodeIndex, 1.f, vec3f(0,0,1));
	}

	if (getType() == BVH_NODETYPE_ROOT ||
		getType() == BVH_NODETYPE_JOINT)
	{
		const BVHGroup *group = (BVHGroup *)this;
		for (int i = 0; i < (int)group->child.size(); i++) {
			vtx_ring_t childRing = group->child[i]->genRing(nodeList, triList, vtxList);
			bindVertexRing(triList, dstRing, childRing);
		}
	}

	return firstRing;
}

//////////////// bvh loading / parsing

class LexemeParser;
class BVHFile;

void BVHFile::load_error(LexemeParser *lex, const char *_format, ...) {
	char errorStr[1024];
	sprintf(errorStr, "line %d col %d - ", lex->lineCount, lex->colCount);
	char *dest = errorStr + strlen(errorStr);

	va_list ap;
	va_start(ap, _format);
	vsprintf(dest, _format, ap);
	va_end(ap);

	printf("error: %s\n", errorStr);
}

void BVHFile::load_expected(LexemeParser *lex, const char *str) {
	load_error(lex, "expected %s, found %s\n", str, lex->lexemeString());
}

//must match, or error
bool BVHFile::load_matches(LexemeParser *lex, const char *str, bool must) {
	if (!strcmp(lex->lexemeString(), str)) {
		lex->process();
		return true;
	}
	if (must) {
		load_expected(lex, str);
	}
	return false;
}

bool BVHFile::load_read_int(LexemeParser *lex, const char *name, int *dest) {
	bool neg = load_matches(lex, "-", false);

	if (lex->token() != TOKEN_CONST_INT) {
		load_expected(lex, name);
		return false;
	}
	dest[0] = atoi(lex->lexemeString());
	if (neg) dest[0] = -dest[0];
	lex->process();
	return true;
}

bool BVHFile::load_read_float(LexemeParser *lex, const char *name, float *dest) {
	bool neg = load_matches(lex, "-", false);

	if (lex->token() != TOKEN_CONST_INT && lex->token() != TOKEN_CONST_FLOAT) {
		load_expected(lex, name);
		return false;
	}
	dest[0] = (float)atof(lex->lexemeString());
	if (neg) dest[0] = -dest[0];
	lex->process();
	return true;
}

bool BVHFile::load_read_string(LexemeParser *lex, const char *name, char *dest, int size) {
	if (lex->token() != TOKEN_CONST_STRING &&
		lex->token() != TOKEN_NAME)
	{
		load_expected(lex, name);
		return false;
	}
	strncpy(dest, lex->lexemeString(), size-1);
	dest[size-1] = 0;
	lex->process();
	return true;
}

bool BVHFile::load_motion(LexemeParser *lex) {
	if (!load_matches(lex, "MOTION")) return false;
	if (!load_matches(lex, "Frames")) return false;
	if (!load_matches(lex, ":")) return false;
	if (!load_read_int(lex, "frame count", &frameCount)) return false;
	if (!load_matches(lex, "Frame")) return false;
	if (!load_matches(lex, "Time")) return false;
	if (!load_matches(lex, ":")) return false;
	if (!load_read_float(lex, "frame time", &frameFreq)) return false;

	//now on to the data:
	//NOTICE - my lexer doesn't discern newlines.  so the numbers better align =)
	//channel data read in from the file
	//this will be sized 'frameCount' by 'channels.size()'
	//it'll also proly be done away with eventually
	int dataSize = frameCount * (int)channels.size();
	float *data = new float[dataSize];
	if (!data) return false;

	int i;
	for (i = 0; i < dataSize; i++) {
		if (!load_read_float(lex, "data var", data+i)) return false;
	}

	//so what if the user put junk at the end of the file?
	//if (lex->token() != TOKEN_EOF) {
	//	load_expected(lex, "end of file");
	//	return false;
	//}

	//now - go through and create our basis data
	basis = new basis34f[frameCount * nodes.size()];
	if (!basis) return false;

	//'b' will be the pointer to the current basis which we are accumulating transforms for
	basis34f *b = basis;
	//dataFrame will be the pointer to the channel data for the current frame
	float *dataFrame = data;

	//for all frames
	for (i = 0; i < frameCount; i++, dataFrame += channels.size()) {

		//for all nodes
		for (int j = 0; j < (int)nodes.size(); j++, b++) {
			BVHNode *node = nodes[j];

			//accumulate the translation/rotation info into this basis

			//store identity rotation
			b->r.set(0,0,0,1);
			//store the offset initially
			b->t = node->offset;

			//see if we have any channels to accumulate for this node:
			if (node->getType() == BVH_NODETYPE_JOINT ||
				node->getType() == BVH_NODETYPE_ROOT)
			{
				BVHGroup *group = (BVHGroup *)node;

				for (int i = group->channelIndex; i < group->channelIndex + group->channelCount; i++) {
					float d = dataFrame[i];

					switch (channels[i]) {
					case BVH_CHANNEL_XPOS:	*b *= basis34f(vec4f(), vec3f(d, 0, 0));	break;
					case BVH_CHANNEL_YPOS:	*b *= basis34f(vec4f(), vec3f(0, d, 0));	break;
					case BVH_CHANNEL_ZPOS:	*b *= basis34f(vec4f(), vec3f(0, 0, d));	break;
					case BVH_CHANNEL_XROT:	*b *= basis34f(angleAxisToQuat(quatf(1, 0, 0, d)), vec3f());	break;
					case BVH_CHANNEL_YROT:	*b *= basis34f(angleAxisToQuat(quatf(0, 1, 0, d)), vec3f());	break;
					case BVH_CHANNEL_ZROT:	*b *= basis34f(angleAxisToQuat(quatf(0, 0, 1, d)), vec3f());	break;
					default:
						printf("found an unknown bvh channel type: %d within node %d\n", channels[i], j);
						break;
					}
				}
			}
		}
	}

	delete data;

	return true;
}

//assert that "ROOT" or "JOINT" or "End Site" has already been read:
BVHNode *BVHFile::load_node(LexemeParser *lex, int type, int parentIndex) {
	char str[256];

	BVHNode *node = NULL;
	switch (type) {
	case BVH_NODETYPE_ROOT:		node = new BVHRoot();		break;
	case BVH_NODETYPE_JOINT:	node = new BVHJoint();		break;
	case BVH_NODETYPE_ENDSITE:	node = new BVHEndSite();	break;
	default:
		load_error(lex, "load_node got an invalid node type: %d\n", type);
		return NULL;
	}
	if (!node) return NULL;

	//store it on the node stack - in the same order that we read it in the file
	node->nodeIndex = (int)nodes.size();
	node->parentIndex = parentIndex;
	nodes.push_back(node);

	//catch the name if it has one
	if (type == BVH_NODETYPE_JOINT ||
		type == BVH_NODETYPE_ROOT)
	{
		if (!load_read_string(lex, "<node name>", str, sizeof(str))) return NULL;
		((BVHGroup *)node)->name = str;
	}

	//expect a "{"
	if (!load_matches(lex, "{")) return NULL;

	//read all our children
	while (lex->token() != TOKEN_BRACE_CLOSE) {

		if (lex->token() == TOKEN_EOF) return NULL;

		//read all the fields within:
		if (load_matches(lex, "OFFSET", false)) {	//false means "if it doesnt match then dont throw an error"

			if (!load_read_float(lex, "<offset_x>", &node->offset.x)) return NULL;
			if (!load_read_float(lex, "<offset_y>", &node->offset.y)) return NULL;
			if (!load_read_float(lex, "<offset_z>", &node->offset.z)) return NULL;

		} else if (type == BVH_NODETYPE_JOINT ||
					type == BVH_NODETYPE_ROOT)
		{
			BVHGroup *group = (BVHGroup *)node;
			
			
			if (load_matches(lex, "CHANNELS", false)) {

				if (!load_read_int(lex, "<channels>", &group->channelCount)) return NULL;
				//now that we have the channel count, allocate a piece of the channel vector and resize accordingly
				group->channelIndex = (int)channels.size();
				channels.resize(channels.size() + group->channelCount);

				//now cycle through and read all the channel types for this node
				for (int i = group->channelIndex; i < group->channelIndex + group->channelCount; i++) {
					if (load_matches(lex, "Xposition", false)) {
						channels[i] = BVH_CHANNEL_XPOS;
					} else if (load_matches(lex, "Yposition", false)) {
						channels[i] = BVH_CHANNEL_YPOS;
					} else if (load_matches(lex, "Zposition", false)) {
						channels[i] = BVH_CHANNEL_ZPOS;
					} else if (load_matches(lex, "Xrotation", false)) {
						channels[i] = BVH_CHANNEL_XROT;
					} else if (load_matches(lex, "Yrotation", false)) {
						channels[i] = BVH_CHANNEL_YROT;
					} else if (load_matches(lex, "Zrotation", false)) {
						channels[i] = BVH_CHANNEL_ZROT;
					} else {
						load_expected(lex, "X|Y|Z position|rotation");
						return NULL;
					}
				}

			} else if (load_matches(lex, "JOINT", false)) {

				BVHGroup *joint = (BVHGroup *)load_node(lex, BVH_NODETYPE_JOINT, node->nodeIndex);
				if (!joint) return NULL;

				group->child.push_back(joint);

			} else if (load_matches(lex, "End", false)
				&& load_matches(lex, "Site"))	//FIXME: in BVH, "End  Site" would be invalid, but my lexer would still accept it
			{
				BVHEndSite *endSite = (BVHEndSite *)load_node(lex, BVH_NODETYPE_ENDSITE, node->nodeIndex);
				if (!endSite) return NULL;

				group->child.push_back(endSite);

			}
		}
	}
	lex->process();
	
	return node;
}

bool BVHFile::load_heirarchy(LexemeParser *lex) {
	if (!load_matches(lex, "HIERARCHY")) return false;
	if (load_matches(lex, "ROOT")) {
		if (!load_node(lex, BVH_NODETYPE_ROOT, -1)) return false;
	}
	return true;
}

#ifdef USE_ISOSURF
static BVHFile *isosurf_bvhfile = NULL;
static IsosurfMesh *isosurf_mesh = 0;
static float isosurf_threshold = 1.f;
static float isosurf_step = 1.5f;
static vec3f isosurf_start;
static float isosurf_func(const vec3f &v) {

	if (!isosurf_bvhfile) return 0;

	float k = 0;

	const std::vector<BVHNode *> &bvhNodes = isosurf_bvhfile->getNodes();

	for (int i = 0; i < bvhNodes.size(); i++) {
		BVHNode *n = bvhNodes[i];
		if (n->parentIndex == -1) continue;

		vec3f a = n->basis.t;
		vec3f b = bvhNodes[n->parentIndex]->basis.t - a;

		//printf("a = %f %f %f\n", VEC3ELEM(a));
		//printf("b = %f %f %f\n", VEC3ELEM(b));

		float bdotb = b % b;
		if (bdotb < 0.01f) continue;

		float t = ((v - a) % b) / bdotb;
		if (t < 0) t = 0;
		if (t > 1) t = 1;

		vec3f x = a + b * t;

		float lensq = vecLengthSq(v - x);
		if (lensq < 0.01f) lensq = 0.01f;
		k += 1.f / lensq;
		//printf("v = %f %f %f\n", VEC3ELEM(v));
		//printf("x = %f %f %f\n", VEC3ELEM(x));
		//printf("lensq = %f\n", lensq);
		//printf("k = %f\n", k);
	}

//	printf("k = %f\n", k);

	return k;
}
#endif


bool BVHFile::init(const char *filename) {
	shutdown();

//// TODO - CLASS-IZE THIS CHUNK OF CODE

	//load the file
	//todo - make me a generic function?

	FILE *file = fopen(filename, "rb");
	if (!file) {
		printf("BVHFile::init - fopen failed\n");
		return false;
	}

	if (fseek(file, 0, SEEK_END)) {
		printf("BVHFile::init - fseek failed\n");
		fclose(file);
		return false;
	}

	int size = ftell(file);
	if (size == -1) {
		printf("BVHFile::init - ftell failed\n");
		fclose(file);
		return false;
	}

	char *buffer = (char *)malloc(size+1);
	if (!buffer) {
		printf("BVHFile::init - malloc failed\n");
		fclose(file);
		return false;
	}
	buffer[size]=0;

	if (fseek(file, 0, SEEK_SET)) {
		printf("BVHFile::init - fseek failed\n");
		fclose(file);
		free(buffer);
		return false;
	}

	if (fread(buffer, size, 1, file) != 1) {
		printf("BVHFile::init - fread failed\n");
		fclose(file);
		free(buffer);
		return false;
	}

	fclose(file);

//// END CHUNK O CODE

	LexemeParser lex(buffer, size);

	bool success = load_heirarchy(&lex);
	if (success) success = load_motion(&lex);

	free(buffer);

#if defined USE_ISOSURF && !defined USE_ISOSURF_REALTIME
	//set up the initial pose for the isosurface function generator:
	setFrame(0);
	genBasis();
	isosurf_bvhfile = this;

	//"Somewhere where the light is strong..."
	//start somewhere where we know the potential function surpasses the threshold
	isosurf_start = nodes[0]->basis.t;

	if (isosurf_mesh ) delete isosurf_mesh ;
	isosurf_mesh = Isosurf_Generate(
		isosurf_func,
		isosurf_threshold,
		isosurf_step,
		isosurf_start);

	if (!isosurf_mesh) {
		printf("isosurf failed\n");
	}
#endif

#ifdef BVH_STORE_HISTORY
	history = new basis34f[nodes.size() * BVH_STORE_HISTORY_COUNT];
	if (!history) {
		printf("failed to init the history\n");
		return false;
	}
#endif //BVH_STORE_HISTORY

#ifdef BVH_DEFORMABLE_MESH
	genIdentityBasis();
	if (nodes[0]) {
		//generate our base model by a collection of interconnected rings
		nodes[0]->genRing(nodes, triList, vtxList);

		//allocate enough space to hold the deformed model (and, potentially, all of its history...)
		int count = vtxList.size();
#ifdef BVH_DEFORMABLE_MESH_HISTORY
		count *= BVH_STORE_HISTORY_COUNT;
#endif //BVH_DEFORMABLE_MESH_HISTORY
		curVtxSet = new vec3f[count];
		if (!curVtxSet) {
			printf("failed to allocate the vertex buffer of size %d\n", count);
			return false;
		}

		curNrmSet = new vec3f[count];
		if (!curNrmSet) {
			printf("failed to allocate the normal buffer of size %d\n", count);
			return false;
		}
	}
#endif

	return success;
}

void BVHFile::shutdown() {
	if (basis) delete basis;

#ifdef BVH_STORE_HISTORY
	if (history) delete history;
#endif

#ifdef BVH_DEFORMABLE_MESH
	triList.clear();
	vtxList.clear();
	if (curVtxSet) delete curVtxSet;
	if (curNrmSet) delete curNrmSet;
#endif

	for (int i = 0; i < (int)nodes.size(); i++) {
		if (nodes[i]) delete nodes[i];
	}

	channels.clear();
	nodes.clear();

	reset();
}

//////////////// rendering

//in my dreams:
//isurface over the skeleton using a lineseg distance-squared function for each bvh graph edge
//upon isurface mesh creation, store the N strongest transform influences associated with each vertex
//then - regen as we retransform

//generates the basii for the current specified 'frame' value
void BVHFile::genBasis() {
	//if we got nothing then draw nothing
	if (!nodes.size()) return;

	//calc this and next frame set, and fraction, for linear interpolation
	int iFrame = (int)frame;
	float frac = frame - (float)iFrame;
	int iNextFrame = iFrame + 1;

	//get the pointers to the beginning of the basis arrays for the current and next frames
	const basis34f *b = basis + (iFrame * nodes.size());			//current basis frame pointer
	const basis34f *b2 = basis + (iNextFrame * nodes.size());	//next basis frame pointer

	//compute all node worldbasis's

	for (int i = 0; i < (int)nodes.size(); i++, b++, b2++) {
		//assert that each node's parent is less than its own node index
		BVHNode *n = nodes[i];

		assert(n->parentIndex < i);
		assert(n->parentIndex >= -1);	//-1 denotes 'no parent'
		
		if (n->parentIndex == -1) {
			//apply to our identity space
			//correct for BVH->my space
			n->basis = basis34f(
#ifndef M_SQRT_1_2
#define M_SQRT_1_2	(float)0.70710678118654752440084436210485
#endif
				quatf(M_SQRT_1_2, 0, 0, M_SQRT_1_2) *
				quatf(0, -M_SQRT_1_2, 0, M_SQRT_1_2),
				vec3f(0,0,0));
		} else {
			//apply to the parent's space
			n->basis = nodes[n->parentIndex]->basis;
		}

		//pull it from the derived basis (which /include/ offset info, mind you.  but dont go removing the offset fields, because we still need them for rendering)
		//calc interpolated values
		n->basis *= basis34f(
			quatSlerp(b->r, b2->r, frac),
			vec3lerp(b->t, b2->t, frac));
	}
}

void BVHFile::genIdentityBasis() {
	//if we got nothing then draw nothing
	if (!nodes.size()) return;

	//compute all node worldbasis's

	for (int i = 0; i < (int)nodes.size(); i++) {
		//assert that each node's parent is less than its own node index
		BVHNode *n = nodes[i];

		assert(n->parentIndex < i);
		assert(n->parentIndex >= -1);	//-1 denotes 'no parent'
		
		if (n->parentIndex == -1) {
			n->basis = basis34f(vec4f(), vec3f());
		} else {
			//apply to the parent's space
			n->basis = nodes[n->parentIndex]->basis;
		}

		//pull it from the derived basis (which /include/ offset info, mind you.  but dont go removing the offset fields, because we still need them for rendering)
		//calc interpolated values
		n->basis *= basis34f(vec4f(), n->offset);
	}
}


void BVHFile::setFrame(float f) {
	frame = ifmod(f, frameCount-1);
}

#ifdef BVH_DEFORMABLE_MESH
void BVHFile::calcDeformMesh() {
	vec3f *vtxFrame = curVtxSet;
	vec3f *nrmFrame = curNrmSet;
#ifdef BVH_DEFORMABLE_MESH_HISTORY
	vtxFrame += history_index * vtxList.size();
	nrmFrame += history_index * vtxList.size();
#endif //BVH_DEFORMABLE_MESH_HISTORY

	//interpolate from nodes[nodes[j].parentIndex].basis to nodes[j].basis by the node index j & coeff in each BVHNode::vtx_t
	vec3f *v = vtxFrame;
	for (int i = 0; i < vtxList.size(); i++, v++) {
		//get the current vertex
		const BVHNode::vtx_t *vtx = &(vtxList[i]);

		//get the node associated with this vertex
		const BVHNode *nodeDst = nodes[vtx->n];
		//get its basis:
		const basis34f *bDst = &nodeDst->basis;
		//see if it has a parent to interpolate back into:
		if (nodeDst->parentIndex == -1) {
			//if not then just use the current node's basis to transform the bone mesh info:
			*v = bDst->t + quatRotate(bDst->r, vtx->x);
		} else {
			//get the parent node
			BVHNode *nodeSrc = nodes[nodeDst->parentIndex];
			//and get its basis
			const basis34f *bSrc = &nodeSrc->basis;
			//now generate based upon the interpolated basis:
			*v = vec_lerp(bSrc->t, bDst->t, vtx->t) + quatRotate(quatSlerp(bSrc->r, bDst->r, vtx->t), vtx->x);
		}
	}

	//gen normals:

	//clear
	for (i = 0; i < vtxList.size(); i++) {
		nrmFrame[i].set(0,0,0);
	}

	//accum
	for (i = 0; i < triList.size(); i+=3) {
		//these may not align - we have some transforms inside our BVH that flips our mesh inside out =(
		//as a result, some normals of vertices on joints may twist around a lot
		const vec3f &a = vtxFrame[triList[i+0]];
		const vec3f &b = vtxFrame[triList[i+1]];
		const vec3f &c = vtxFrame[triList[i+2]];
		vec3f n = vec3unit((b - a) * (c - a));
		nrmFrame[triList[i+0]] += n;
		nrmFrame[triList[i+1]] += n;
		nrmFrame[triList[i+2]] += n;
	}

	//normalize
	for (i = 0; i < vtxList.size(); i++) {
		nrmFrame[i] = vec3unit(nrmFrame[i]);
	}
}
#endif

#ifdef BVH_STORE_HISTORY
void BVHFile::storeHistory(float dt) {
	//copy over all our new basii into our history (todo - unify later?)
	basis34f *historyLastFrame = history + ((history_index + 1) % BVH_STORE_HISTORY_COUNT) * nodes.size();
	basis34f *historyFrame = history + history_index * nodes.size();
	for (int i = 0; i < nodes.size(); i++) {
		historyFrame[i].t = nodes[i]->basis.t;


		historyFrame[i].r = 
//zero				//vec4f();
//random			//frand()-0.5f,frand()-0.5f,frand()-0.5f,0)*0.1f;
//velocity approx:	//vec4f((historyFrame[i].t - historyLastFrame[i].t) * 0.01f, 0);
//outwards:
				vec4f(historyFrame[i].t - historyLastFrame[i].t,0);
		if (nodes[i]->parentIndex >= 0) {
			*historyFrame[i].r.vp() = *historyFrame[i].r.vp() * (historyFrame[i].t - historyFrame[nodes[i]->parentIndex].t);
		}
		*historyFrame[i].r.vp() = vec3unit(*historyFrame[i].r.vp());
	}

	for (i = 0; i < BVH_STORE_HISTORY_COUNT * nodes.size(); i++) {
		history[i].t += *history[i].r.vp() * dt;
	}
}
#endif

void BVHFile::update(float dt) {

	//convert time to frame number
	float df = dt / frameFreq;

#ifdef BVH_STORE_HISTORY
	float lastFrame = frame;
#endif

	//set the new frame:
	setFrame(frame + df);

	//increment the history index:
#ifdef BVH_STORE_HISTORY
	float motion_dt = history_freq / (float)BVH_STORE_HISTORY_COUNT;
	//if there are any mods of motion_trail_life between system_thistime and system_lasttime... make an image
	if ((int)(frame / motion_dt) != (int)(lastFrame / motion_dt)) {
		//decrement so that our last history write will be stored
		history_index += BVH_STORE_HISTORY_COUNT - 1;
		history_index %= BVH_STORE_HISTORY_COUNT;
	}
#endif

	genBasis();

#ifdef BVH_DEFORMABLE_MESH
	calcDeformMesh();
#endif


#ifdef BVH_STORE_HISTORY
	storeHistory(dt);
#endif

#if defined USE_ISOSURF && defined USE_ISOSURF_REALTIME
	isosurf_bvhfile = this;

	//"Somewhere where the light is strong..."
	//start somewhere where we know the potential function surpasses the threshold
	isosurf_start = nodes[0]->basis.t;

	if (isosurf_mesh ) delete isosurf_mesh ;
	isosurf_mesh = Isosurf_Generate(
		isosurf_func,
		isosurf_threshold,
		isosurf_step,
		isosurf_start);

	if (!isosurf_mesh) {
		printf("isosurf failed\n");
	}
#endif
}

void BVHFile::render(const View *view) {

	//render all data:

	//if we got nothing then draw nothing
	if (!nodes.size()) return;

#if 0	//draw a wireframe
	glLineWidth(5.f);
	glBegin(GL_LINES);
	for (int i = 0; i < nodes.size(); i++) {
		BVHNode *n = nodes[i];
		if (n->parentIndex == -1) continue;

		glVertex3fv(nodes[n->parentIndex]->basis.t.ptr());
		glVertex3fv(n->basis.t.ptr());
	}
	glEnd();
	glLineWidth(1.f);
#endif

#ifdef USE_ISOSURF
	if (isosurf_mesh) {
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		Texture::unbind();
		glColor3f(1,1,1);
		glBegin(GL_TRIANGLES);
		for (int i = 0; i < isosurf_mesh->tris.size(); i++) {
			glVertex3fv(isosurf_mesh->vtxs[isosurf_mesh->tris[i]].ptr());
		}
		glEnd();
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
#endif


	if (skyTex) {

		glMatrixMode(GL_TEXTURE);		//init cubemap reflection texture matrix
		glPushMatrix();					//assume identity is already in the texmat
		view->correctAngleInverse();	//correct from GL space to our space
		view->rotateInverse();			//do a cam inverse rotation
		glRotatef(-90, 1, 0, 0);		//correct into skymap space
		glMatrixMode(GL_MODELVIEW);

		skyTex->bind();

		//init cubemap texcoord generators
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
		glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);

		//enable texcoords
		glEnable(GL_TEXTURE_GEN_S);
		glEnable(GL_TEXTURE_GEN_T);
		glEnable(GL_TEXTURE_GEN_R);
	} else {
		Texture::unbind();
	}
	glBlendFunc(GL_ONE, GL_ONE);		//GL_DST_COLOR, GL_ZERO);
	glDisable(GL_CULL_FACE);
	glEnable(GL_NORMALIZE);

#if defined BVH_DEFORMABLE_MESH
	glColor3f(1,1,1);
	{
		vec3f *vtxFrame = curVtxSet;
		vec3f *nrmFrame = curNrmSet;
#ifdef BVH_DEFORMABLE_MESH_HISTORY
		vtxFrame += history_index * vtxList.size();
		nrmFrame += history_index * vtxList.size();
#endif //BVH_DEFORMABLE_MESH_HISTORY

		glBegin(GL_TRIANGLES);
		for (int i = 0; i < triList.size(); i++) {
			int index = triList[i];
			glNormal3fv(nrmFrame[index].ptr());
			glVertex3fv(vtxFrame[index].ptr());
		}
		glEnd();
	}
#endif //BVH_DEFORMABLE_MESH 


#if defined BVH_STORE_HISTORY

	glEnable(GL_BLEND);
	glDepthMask(0);
	glBegin(GL_QUADS);

	for (int i = 0; i < BVH_STORE_HISTORY_COUNT-1; i++) {

		float f =  (float)i / (float)(BVH_STORE_HISTORY_COUNT-1);
		f *= f*f*f;
		f = 1.f - f;
		f *= 0.5f;
//		f = 1.f - f;
		glColor3f(f,f,f);

		int r = (i + history_index) % BVH_STORE_HISTORY_COUNT;

#if 0 && defined BVH_DEFORMABLE_MESH && defined BVH_DEFORMABLE_MESH_HISTORY

		vec3f *vtxFrame = curVtxSet + r * vtxList.size();
		vec3f *nrmFrame = curNrmSet + r * vtxList.size();

		for (int j = 0; j < triList.size(); j++) {
			int index = triList[j];
			glNormal3fv(nrmFrame[index].ptr());
			glVertex3fv(vtxFrame[index].ptr());
		}
#else
		int r2 = (r + 1) % BVH_STORE_HISTORY_COUNT;

		basis34f *historyFrame = history + r * nodes.size();
		basis34f *historyNextFrame = history + r2 * nodes.size();

		//render from r to r2
		for (int n = 0; n < nodes.size(); n++) {
			int pn = nodes[n]->parentIndex;
			if (pn == -1) continue;

			//render from history[pn] to historyNext[n]
			glNormal3fv(historyFrame[n].r.ptr());
			glVertex3fv(historyFrame[n].t.ptr());
			glNormal3fv(historyFrame[pn].r.ptr());
			glVertex3fv(historyFrame[pn].t.ptr());
			glNormal3fv(historyNextFrame[pn].r.ptr());
			glVertex3fv(historyNextFrame[pn].t.ptr());
			glNormal3fv(historyNextFrame[n].r.ptr());
			glVertex3fv(historyNextFrame[n].t.ptr());
			
		}
#endif
	}

	glEnd();

#endif //BVH_STORE_HISTORY
	
	glDisable(GL_BLEND);

	if (skyTex) {
		//disable texcoords
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_TEXTURE_GEN_R);

		//disable cube mapping
		Texture::unbind();

		//disable cubemap reflection texture matrix
		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);

		//disable texture environment
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	}

	glDepthMask(1);
	glEnable(GL_CULL_FACE);
	glColor3f(1,1,1);

}

