#include <assert.h>
#include "lib/cloth.h"

void Cloth::clear() {
	SpringObj::clear();
	level = 0;
	bbox.min.set(0,0,0);
	bbox.max.set(0,0,0);
}

bool Cloth::createSpring(int e, Particle *v0, Particle *v1) {
	assert(v0);
	assert(v1);
	assert(springSys);
	assert(parSys);

	spring[e] = springSys->createSpring();
	if (spring[e] == -1) return false;

	Spring &s = springSys->spring[spring[e]];
	s.par[0] = v0;
	s.par[1] = v1;
	s.ks = 360.f;
	s.kd = 3.6f;
	s.restlen = 0;

	return true;
}

bool Cloth::init(int level, ParSystem *parSys, SpringSystem *springSys, const char *filename) {
	shutdown();

	if (!SpringObj::init(parSys, springSys)) return false;

	this->level = level;

	if (!tex.load2D(filename)) {
		printf("Cloth.init() failed to load texture filename %s\n", filename);
		return false;
	}

	vertex.resize(level * level);
	spring.resize(6 * level * level - 10 * level + 2);

	int i,j,e = 0;

	for (j = 0; j < level; j++) {
		for (i = 0; i < level; i++)	{
			vertex[e] = parSys->createParticle();
			if (!vertex[e]) return false;
			Particle *p = vertex[e];
			p->massInv = 1;
			p->x.set(
				(float)i / (float)(level-1) - 0.5f,
				(float)j / (float)(level-1) - 0.5f, 0.f);
			p->v.set(0,0,0);
			e++;
		}
	}

	e = 0;
	for (i = 0; i < level; i++) {
		for (j = 0; j < level; j++) {
			if (i > 0)				  {	if (!createSpring(e, vertex[i+j*level], vertex[i-1+j*level]		)) return false; e++; }
			if (j > 0)				  {	if (!createSpring(e, vertex[i+j*level], vertex[i+(j-1)*level]	)) return false; e++; }
			if (i > 0 && j > 0)		  {	if (!createSpring(e, vertex[i+j*level], vertex[i-1+(j-1)*level]	)) return false; e++; }
			if (i > 0 && j < level-1) {	if (!createSpring(e, vertex[i+j*level], vertex[i-1+(j+1)*level]	)) return false; e++; }
			if (i > 1)				  {	if (!createSpring(e, vertex[i+j*level], vertex[i-2+j*level]		)) return false; e++; }
			if (j > 1)				  {	if (!createSpring(e, vertex[i+j*level], vertex[i+(j-2)*level]	)) return false; e++; }
		}
	}

	resetSpringLengths();
	updateBBox();

	return true;
}

void Cloth::render() {
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_CULL_FACE);
	glColor3f(1,1,1);
	tex.bind();
	float il = 1.f / (float)(level - 1);
	for (int j = 0, e = 0; j < level-1; j++) {
		glBegin(GL_TRIANGLE_STRIP);
		for (int i = 0; i < level; i++, e++) {
			glTexCoord2f( (float)i * il, (float)(j+1) * il);
			glVertex3fv(vertex[e+level]->x);
			glTexCoord2f( (float)i * il, (float)j * il);
			glVertex3fv(vertex[e]->x);
		}
		glEnd();
	}
	glPopAttrib();
}

void Cloth::lockEdgeVertexes() {
	for (int i = 0; i < level-1; i++) {
		//top
		vertex[i]->massInv = 0;
		//bottom
		vertex[i+1+(level-1)*level]->massInv = 0;
		//left
		vertex[(i+1)*level]->massInv = 0;
		//right
		vertex[level-1+i*level]->massInv = 0;
	}
}

void Cloth::getTri(int i, Particle **v) {
	int u = i & 1;			//let the first bit be the up/down bit
	i >>= 1;				//shift the bit
	int y = i / (level - 1);		//divide by x range to get y value
	int x = i - y * (level - 1);	//subtract out y influence to get x value
	i = x + y * level;
	//i is now <x,y> ranged <0,0>-><level-1,level-2>
	//for upper:
	//	tri = [x,y+1], [x,y], [x+1,y+1]
	//for lower:
	//	tri = [x,y] [x+1,y+1], [x+1,y]
	if (u) {
		v[0] = vertex[i+level];
		v[1] = vertex[i];
		v[2] = vertex[i+level+1];
	} else {
		v[0] = vertex[i];
		v[1] = vertex[i+level+1];
		v[2] = vertex[i+1];
	}
}

void Cloth::updateBBox() {
	assert(parSys);

	if (!vertex.size()) {
		//todo - fill with -+inf <=> void volume?
		bbox.min = bbox.max = vec3f(0,0,0);
		return;
	}

	//start with the first vertex
	bbox.min = bbox.max = vertex[0]->x;
	//and stretch to include all subsequent vertices
	for (int i = 1; i < (int)vertex.size(); i++) {
		const Particle *p = vertex[i];
		bbox.stretch(p->x);	
	}
}