#include <stdio.h>

#include "lib/collision.h"

float testSphereTriCollision(
	const vec3f &p1, const vec3f &p2, float radius, const vec3f **points, plane3f *resultPlane, float *bcc)
{

	//give different biases to different collision types
	//promote face collisions over others
		//used with stepback
	const float FACE_STEP_EPSILON = 0.03125f;	//0.0000003125;		//keep this very VERY small or else tracelines over far distances will fail

	vec3f delta = p2 - p1;

	//first do a plane test
	//if it collides within the triangle surface bounds then use it
	{
		plane3f plane = planeBuildUnit(*points[0], *points[1], *points[2]);

		if (dot(delta, plane.normal()) <= 0)
		{	//we're flying into it...

			// point in space: a=x+vt for 0<=t<=1
			// point on plane: -r <= (a-p).n <= r
			// substitute:	-r <= (x+vt-p).n <= r
			//	solution:		t=[-p.n-x.n +- r] / (v.n)
			//
			//calculate 't'.  If t < 0 || t > T then throw it away.  else use it as the collision function

			//delta = p2 - p1
			//delta dot normal < 0
			//therefore all numerator values will negatively affect the result value

			//t1 is the plane distance to sphere surface entry collision
			//t1hi is the plane distance to sphere surface exit collision

			float indotd = 1.f / dot((vec3f&)plane, delta);

			float t1 = (radius - plane.w - dot(p1, (vec3f&)plane)) * indotd;
			float t1hi = t1 - 2.f * radius * indotd;
			float t2 = (radius - plane.w - dot(p2, (vec3f&)plane)) * indotd;

//			printf(" t: %f %f\n", t1, t2);

//			//the line start & end lie completely on the front or back - don't bother test this triangle
			if (t1 < 0.f && t2 < 0.f) {
				//for all-on-the-back-side, add an extra test to make sure the back of our sphere also clears the plane:
				if (t1hi < 0) return 1.f;

				//and if not - then just assume we are colliding at the current point:
				t1 = 0;
			}

			if (t1 >= 0.f && t2 >= 0.f) {
				return 1.f;
			}

			//calculate the fraction of movement til we hit the tempP4d
			//adjust for epsilon to keep from flying too close to walls
			float frac;
			if (t1 < 0) {
				frac = (t1 + FACE_STEP_EPSILON) / (t1 - t2);
			} else {
				frac = (t1 - FACE_STEP_EPSILON) / (t1 - t2);
			}

			//clamp adjusted value to [0,1]
			if (frac < 0.f) frac = 0.f;
			if (frac > 1.f) frac = 1.f;

			//calculate collision point
			vec3f mid = vec3lerp(p1, p2, frac);

			//printf("mid dist: %f\n", planePointDist(plane, mid));
			//printf("mid: %f %f %f\n", VEC3ELEM(mid));

			//since 'mid' is a good 16 units off the tempP4d
			//then maybe this line is pretty necessary?
			mid = planeProjectPoint(plane, mid);

			//calculate the triangle basis
			vec3f p0ToMid = mid - *points[0];
			vec3f basis[2] = {
				*points[1] - *points[0],
				*points[2] - *points[0]
			};

////first way to solve:		these worked before, but at the moment im getting strange results...
//				//solve for the barycentric coordinates - store in 'barycentricCoords'
//				//calculate matrix determinant
//				double dot0d = basis[0].dot(p0ToMid);
//				double dot1d = basis[1].dot(p0ToMid);
//				double dot00 = basis[0].lengthSquared();
//				double dot01 = basis[0].dot(basis[1]);
//				double dot11 = basis[1].lengthSquared();
//				double invDet = dot00 * dot11 - dot01 * dot01;
//				if (invDet > 0.0) {
//					invDet = 1.0 / invDet;
//				}
//				double[] bcc = new double[3];
//				//solve the inverse matrix
//				bcc[1] = (dot11 * dot0d - dot01 * dot1d) * invDet;
//				bcc[2] = (dot00 * dot1d - dot01 * dot0d) * invDet;
////second way to solve: the math looks the same (just with a vector quad product identity rewrite)
			vec3f dxd1 = cross(p0ToMid, basis[0]);
			vec3f dxd2 = cross(p0ToMid, basis[1]);
			vec3f d1xd2 = cross(basis[0], basis[1]);
			float div = 1.f / d1xd2.lenSq();

			bcc[1] = dot(dxd2, d1xd2) * div;
			bcc[2] = -dot(dxd1, d1xd2) * div;
////end solve methods
			//compute the third variable
			bcc[0] = 1.f - bcc[1] - bcc[2];

			//calc determinant of the points to find the triangle area
			float twoTriArea =  (float)d1xd2.len();

			//calc the length of each base vector
			float bcEdgeLen[3];
			bcEdgeLen[0] = twoTriArea / (float)(*points[1] - *points[2]).len();
			bcEdgeLen[1] = twoTriArea / (float)(*points[2] - *points[0]).len();
			bcEdgeLen[2] = twoTriArea / (float)(*points[0] - *points[1]).len();

			//tempP4d collision test

			bool collided = false;
				//within all edges inner bounds - collision imminant
			if (bcc[0] > 0 &&
				bcc[1] > 0 &&
				bcc[2] > 0)
			{
				collided = true;

				//outside all edges outer bounds - no collision at all
			} else if (bcc[0] * bcEdgeLen[0] < -radius ||
					bcc[1] * bcEdgeLen[1] < -radius ||
					bcc[2] * bcEdgeLen[2] < -radius)
			{
				collided = false;

				//test each edge to see if we're at the corner & need a vertex test
			} else {
				int i,j,k;
				for (j = 0; j < 3; j++) {
					i = (j + 2) % 3;
					k = (j + 1) % 3;

					vec3f pd = *points[i] - *points[k];	//use basis?
					vec3f vd = mid - *points[i];

					//first for pi:
					if (dot(pd, vd) > 0) {
						//then do a vertex test at pi
						if (dot(vd, vd) < radius*radius) {	//then we collided
							collided = true;
						} else {					//we didn't collide
							collided = false;
						}
						break;
					}

					vd = mid - *points[k];

					if (dot(pd, vd) < 0) {
						//then do a vertex test at pk
						if (dot(vd, vd) < radius*radius) {	//then we collided
							collided = true;
						} else {				// we didn't collide
							collided = false;
						}
						break;
					}

					//do an edge test along i->k
					if (bcc[j] * bcEdgeLen[j] > -radius) {
						collided = true;
						break;
					}

					//this final test should be irrelevant because, in theory, 
					//the first two tests that set collision= should have already
					//tested this condition over the whole triangle. there's no point
					//in testing it over just this edge

					collided = false;
					//System.out.println("bcc"+j+"="+bcc[j]+" > epsilon="+(-radius/bcEdgeLen[j]));
					break;
				}
			}

			////test collisions to the exact area of the triangle
			if (collided) {
				*resultPlane = plane;		//todo - normalize the point
//				printf("col on plane %f %f %f %f\n", VEC4ELEM(plane));
				return frac;
			}
		}
	}

	return 1.f;
}

//this isn't working so hot for our spring system - we're getting stuck objects due to not collision testing cloth particle movement
float testSphereSphereCollision(
	const vec3f &p1,
	const vec3f &p2,
	float radius,
	const vec3f &pTest,
	float radTest,
	plane3f *resultPlane)
{
	//p = (p1 + (p2-p1)t) - pTest
	//r = radius + radTest
	//collision when p.p <= r*r
	//p.p = ((p1 + (p2-p1)t) - pTest).((p1 + (p2-p1)t) - pTest)
	//p.p = ((p1 - pTest) + (p2-p1)t).((p1 - pTest) + (p2-p1)t)
	//let a = p1 - pTest
	//		b = p2 - p1
	//p.p = (a + b*t)^2
	//p.p = a.a + 2a.bt + b.bt^2
	//r*r < a.a + 2a.bt + b.bt^2
	//0 < a.a-r*r + 2a.bt + b.bt^2
	//t = [-2a.b +- sqrt[(2a.b)^2 - 4(a.a-r*r)*b.b] ] / [2b.b]
	//t = [-a.b +- sqrt[(a.b)^2 - (a.a-r*r)*b.b] ] / [b.b]

	vec3f a = p1 - pTest;
	vec3f b = p2 - p1;
	float r = radTest + radius;
	float rr = r * r;
	float ab = dot(a, b);
	float aa = dot(a, a);
	float bb = dot(b, b);
	float in = ab*ab - (aa-rr)*bb;
	if (in < 0) return 1.f;	//we never touch
	float sqrtin = sqrtf(in);
	float t0 = -ab - sqrtin;
	float t1 = -ab + sqrtin;

	//no solution:
	//  |  \_/
	//  |  
	//--0--------1
	//
	//one solution:
	//  |  |   |
	//  |   \ /
	//--0----*---1
	//		t0=t1
	//
	//two solutions:
	//  |  |   |
	//  |  |   |
	//--0--*---*-1
	//      \_/
	//	   t0<t1

	//got a collision!
	if (t1 < 0) return 1.f;		//we've passed the object
	if (t0 > bb) return 1.f;	//we will eventually hit it, but not yet
	float t = 0.f;				//first assume we are colliding
	if (t0 > 0) t = t0;			//but push our time back if we havent hit it immediately

	t /= bb;					//rescale our timestep from [0,b.b] to [0,1]
	vec3f colx = vec3lerp(p1, p2, t);
	vec3f del = colx - pTest;
	vec3f mid = pTest + del * (radTest / (radTest + radius));

	(vec3f&)*resultPlane = vec3unit(del);
	resultPlane->calcDist(mid);

	return t;
}
