#include <stdio.h>
#include <stdlib.h>

#include "defs/defs.h"
#include "lib/landscape.h"
#include "lib/fractal.h"

void Landscape::reset() {
	vtx = NULL;
	detail = 0;
	scale.set(1,1,1);
	translate.set(0,0,0);
	texDetailX = 100;
	texDetailY = 100;
	maxvalue = 1;
	minvalue = 0;
	list_id = 0;

	restitution = 0.5f;
	friction = 0.75f;
}

bool Landscape::init(int detail, int filter_level, int seed) {
	shutdown();

	this->scale = scale;
	this->detail = detail;

	srand(seed);rand();rand();rand();

	//init
	vtx = new vtx_t[detail*detail];
	if (!vtx) {
		printf("failed to alloc the vertices\n");
		return false;
	}

	//fractal-generate
	if (!Fractal_Create(
		&vtx->height,	//float *data,
		0,				//float min,
		1,				//float max,
		detail,			//int width,
		detail,			//int height,
		sizeof(vtx_t),	//int step_size,
		filter_level))	//int filter_level);
	{
		printf("fractal landscape creation failed\n");
		return false;
	}

	updateMesh();

	return true;
}

void Landscape::shutdown() {
	if (vtx) delete vtx;
	if (list_id) {
		glDeleteLists(list_id, 1);
	}
	//unalloc
	reset();
}

void Landscape::updateMesh() {
	vec3f lightVector = vec3unit(vec3f(1,-1,2));
	float ambientLightLevel = 0.4f;

	//the first compare will pass
	maxvalue = logf(0);
	minvalue = -logf(0);

	int i,j;
	vtx_t *v = vtx;
	for (j = 0; j < detail; j++) {
		for (i = 0; i < detail; i++, v++) {

			if (v->height < minvalue) minvalue = v->height;
			if (v->height > maxvalue) maxvalue = v->height;

			static vec3f altcolors[] = {
				vec3f(0.4f, 1.f, 0.1f),
				vec3f(0.2f, 0.4f, 0.2f),
				vec3f(0.25f, 0.6f, 0.3f),
				vec3f(0.3f, 0.5f, 0.2f),
				vec3f(0.4f, 1.f, 0.3f),
				vec3f(0.6f, 0.8f, 0.5f),
				vec3f(0.6f, 0.7f, 0.8f),
				vec3f(1,1,1)
			};
			//height is now ranged 0 to 1
			float h = v->height;
			//permute and clamp
			h += frand() * 0.1f - 0.05f;
			if (h < 0) h = 0;
			if (h > 1) h = 1;

			//calc our scaled height by our discretized unit count
			float adjh = h * (float)(numberof(altcolors)-1);
			//find our discretized value
			int disc = (int)(adjh);
			//calc the fraction of movement into this discretized value
			float frac = adjh - (float)disc;
			//interpolate colorspace
			if (disc == numberof(altcolors)-1) {
				v->color = altcolors[disc];
			} else {
				v->color = vec3lerp( altcolors[disc], altcolors[disc+1], frac);
			}

			//finally, cross normals to find light vector and dot for pre-compted diffuse shading
			vec3f dx(
				scale.x * 2.f / (float)(detail-1),
				0,
				scale.z * (vtx[ ((i+1)%detail) + j*detail].height - vtx[ ((i+detail-1)%detail) + j*detail].height));
			vec3f dy(
				0,
				scale.y * 2.f / (float)(detail-1),
				scale.z * (vtx[ i + ((j+1)%detail)*detail].height - vtx[ i + ((j+detail-1)%detail)*detail].height));
			vec3f cr = vec3unit(cross(dx, dy));
			float dt = dot(cr, lightVector);
			if (dt < ambientLightLevel) dt = ambientLightLevel;
			v->color *= dt;

		}
	}
}

void Landscape::renderImmediateData() {
	int i,j;
	vtx_t *v = vtx;
	float id = 2.f / (float)(detail-1);
	float x,y;
	for (j = 0, y = -1.f; j < detail-1; j++, y+=id) {
		glBegin(GL_TRIANGLE_STRIP);
		for (i = 0, x = -1.f; i < detail; i++, x+=id, v++) {
			glColor3fv(v[detail].color);
			glVertex3f(x,y+id,v[detail].height);

			glColor3fv(v[0].color);
			glVertex3f(x,y,v[0].height);
		}
		glEnd();
	}
}

void Landscape::render() {

	glPushMatrix();
	glTranslatef(translate.x, translate.y, translate.z);
	glScalef(scale.x, scale.y, scale.z);

	tex.bind();

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGenfv(GL_S, GL_OBJECT_PLANE, vec4f(texDetailX, 0, 0, 0));
	glTexGenfv(GL_T, GL_OBJECT_PLANE, vec4f(0, texDetailY, 0, 0));
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);

	//apply our detail map:
	if (found_GL_ARB_multitexture) {
		glActiveTextureARB(GL_TEXTURE1_ARB);
		tex.bind();
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);	//should be default, but just in case
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
		glTexGenfv(GL_S, GL_OBJECT_PLANE, vec4f(8.f * texDetailX, 0, 0, 0));
		glTexGenfv(GL_T, GL_OBJECT_PLANE, vec4f(0, 8.f * texDetailY, 0, 0));
		glEnable(GL_TEXTURE_GEN_S);
		glEnable(GL_TEXTURE_GEN_T);
	}

	if (list_id) {
		glCallList(list_id);
	} else {
		renderImmediateData();
	}

	//disable detail map
	if (found_GL_ARB_multitexture) {
		glTexGenfv(GL_S, GL_OBJECT_PLANE, vec4f(1, 0, 0, 0));
		glTexGenfv(GL_T, GL_OBJECT_PLANE, vec4f(0, 1, 0, 0));
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		Texture::unbind();
		glActiveTextureARB(GL_TEXTURE0_ARB);
	}

	//reset our texture gen stuff:
	glTexGenfv(GL_S, GL_OBJECT_PLANE, vec4f(1, 0, 0, 0));
	glTexGenfv(GL_T, GL_OBJECT_PLANE, vec4f(0, 1, 0, 0));
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);

	glPopMatrix();

}

void Landscape::gaussFalloff(const vec3f &src) {
	//v is in world coordinates.  Landscape:: will have to inverse-transform them by its basis
	vec3f iv = src;
	iv -= translate;
	iv.x /= scale.x;
	iv.y /= scale.y;
	iv.z /= scale.z;

	//now do our gauss stuff

//	printf("inv coords: %f %f %f\n", VEC3ELEM(iv));

	int i,j;
	vtx_t *v = vtx;
	float id = 2.f / (float)(detail-1);
	float x,y;
	for (j = 0, y = -1.f; j < detail; j++, y+=id) {
		for (i = 0, x = -1.f; i < detail; i++, x+=id, v++) {
			//calc delta in landscape space
			float dx = x - iv.x;
			float dy = y - iv.y;
			float lensq = dx*dx + dy*dy;
			//do our gauss curve
			float conv = exp(-lensq * 3.f);
			//and pull to the desired height according to our coefficient
			v->height += (iv.z - v->height) * conv;
		}
	}
}

void Landscape::compile() {
	if (list_id) {
		glDeleteLists(list_id, 1);
	}

	list_id = glGenLists(1);
	glNewList(list_id, GL_COMPILE);
	renderImmediateData();
	glEndList();
}

bool Landscape::testSphereCollision(const vec3f &src, float rad, col_t *col) {
	box3f lscolvol(src);
	lscolvol.stretch(col->x);
	lscolvol.min -= vec3f(1,1,1);
	lscolvol.max += vec3f(1,1,1);

	//retranslate & rescale
	lscolvol.min -= translate;
	lscolvol.max -= translate;
	lscolvol.min = vec3InvScaleElem(lscolvol.min, scale);
	lscolvol.max = vec3InvScaleElem(lscolvol.max, scale);

	//landscpae space range
	static box3f bbox( vec3f(-1,-1,minvalue), vec3f(1,1,maxvalue));

	if (!lscolvol.touches(bbox)) return false;

	//inclusive range
	int txmin = (int)((lscolvol.min.x + 1.f) * 0.5f * (float)(detail-1));
	int txmax = (int)((lscolvol.max.x + 1.f) * 0.5f * (float)(detail-1))+1;
	int tymin = (int)((lscolvol.min.y + 1.f) * 0.5f * (float)(detail-1));
	int tymax = (int)((lscolvol.max.y + 1.f) * 0.5f * (float)(detail-1))+1;

	//just in case we got something /right on/ the edge
	if (txmin < 0)
		txmin = 0;
	if (txmax >= detail)
		txmax = detail-1;
	if (tymin < 0)
		tymin = 0;
	if (tymax >= detail)
		tymax = detail-1;

	//see if we can test the children:

	bool collided = false;

	float vwdx = scale.x * 2.f / (float)(detail-1);
	float vwdy = scale.y * 2.f / (float)(detail-1);

	float vwx, vwy;

	int i,j;
	for (j = tymin, vwy = translate.y - scale.y + vwdy * (float)tymin; j <= tymax; j++, vwy += vwdy) {
		for (i = txmin, vwx = translate.x - scale.x + vwdx * (float)txmin; i <= txmax; i++, vwx += vwdx) {
			vtx_t *v = &vtx[i + j * detail];

			vec3f v00(vwx, vwy, v->height * scale.z + translate.z);
			vec3f v10(vwx + vwdx, vwy, v[1].height * scale.z + translate.z);
			vec3f v01(vwx, vwy + vwdy, v[detail].height * scale.z + translate.z);
			vec3f v11(vwx + vwdx, vwy + vwdy, v[detail+1].height * scale.z + translate.z);

			for (int flip = 0; flip < 2; flip++) {	//test both tris in this quad

				const vec3f *pt[3];

				switch(flip) {
				case 0: pt[0] = &v01; pt[1] = &v00; pt[2] = &v11; break;
				case 1: pt[0] = &v10; pt[1] = &v11; pt[2] = &v00; break;
				}

				//test for a collision with that triangle
				float bcc[3];
				float frac = testSphereTriCollision(src, col->x, rad, pt, &col->p, bcc);

				if (frac < 1.f) {
					collided = true;

					col->vtx[0] = *pt[0];
					col->vtx[1] = *pt[1];
					col->vtx[2] = *pt[2];
					
					col->bcc[0] = bcc[0];
					col->bcc[1] = bcc[1];
					col->bcc[2] = bcc[2];

					//landscape surface values
					col->restitution = restitution;
					col->friction = friction;

					if (frac == 0.f) {
						col->x = src;
						col->frac = 0;
						return true;			//we've stopped - no more collisions could occur =)
					}
					col->x = vec3lerp(src, col->x, frac);
					col->frac *= frac;
				}
			}
		}
	}

	return collided;
}
