#include <stdio.h>
#include <assert.h>
#include "lib/objinst.h"

void Material::setup(int face) {
	glMaterialfv(face, GL_AMBIENT, amb);
	glMaterialfv(face, GL_DIFFUSE, diff);
	glMaterialfv(face, GL_SPECULAR, spec);
	//todo - get ahold of nv_max_shininess, or whatever that extension is, for greater range on shininess
	glMaterialf(face, GL_SHININESS, shine);
}

static int objIndex = 0;

void ObjInst::reset() {
	model = NULL;
	triextra = NULL;
	vtxextra = NULL;
	bspnode = NULL;
	visCount = 0;

	//rendering
	cubeTex.unload();
	mat.diff = vec4f(1,1,1,1);
	reflect = 0.5f;
 
	//transformations
	basis.t = vec3f(0,0,0);
	basis.r = quat4fIdentity;
	scale = vec3f(1,1,1);

	bbox.min.set(0,0,0);
	bbox.max.set(0,0,0);

	//collision
	restitution = 0.5f;
	friction = 0.75f;

	//title
	sprintf(title, "obj %d\n", objIndex++);
}

void ObjInst::unload() {
	if (triextra) delete triextra;
	if (vtxextra) delete vtxextra;
	if (bspnode) delete bspnode;
	reset();
}

void ObjInst::render() {
	if (!model) return;

#if 1

	//bind the object instance's texture
	tex.bind();

	//and transform to the object's local space
	glPushMatrix();

	glTranslatef(basis.t.x, basis.t.y, basis.t.z);
	vec4f aa = quatToAngleAxis(basis.r);
	glRotatef(aa.w, aa.x, aa.y, aa.z);
	glScalef(scale.x, scale.y, scale.z);
	model->renderImmediate(false);

	glPopMatrix();

#else	//all below is something to do with shadow casting... an old old now unsupported feature...

#if 1
	model->render();
#else

	const std::vector<tri_t> &tri = model->getTris();
	const std::vector<vec3f> &vertex = model->getVertexes();
	const std::vector<edge_t> &edge = model->getEdges();

#if 0	//edge rendering

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_DEPTH_TEST);
	glLineWidth(5.f);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	glColor3f(1, 0.5f, 0.25f);
	glBegin(GL_LINES);
	for (int ei = 0; ei < edge.size(); ei++) {
		const edge_t &e = edge[ei];

		//get our two tri info's
		int i = e.t[0];
		int j = e.t[1];

		int k = e.v[0];
		int kk = (k + 1) % 3;
		int l = e.v[1];
		int ll = (l + 1) % 3;

		const vec3f &v1 = vertex[tri[i].p[kk].v];
		const vec3f &v2 = vertex[tri[i].p[k].v];
		glVertex3fv(v1.v);
		glVertex3fv(v2.v);
	}
	glEnd();
	glPopAttrib();
#endif

#if 0	//software gouraud lighting
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glColor3f(1, 0.5f, 0.25f);
	glBegin(GL_TRIANGLES);
	for (int i = 0; i < tri.size(); i++) {
		for (int j = 0; j < 3; j++) {

			const vec3f &vtx = vertex[ tri[i].p[j].v ];
			vec3f lightvec = -(quatRotate(basis.r, vtx) + basis.t);
			vec3f normal = quatRotate(basis.r, tri[i].plane.normal);
			float dot3 = (normal % lightvec) / vecLength(lightvec);
			glColor3f(	color.x * dot3,
						color.y * dot3,
						color.z * dot3);
			glVertex3fv(vtx.v);
		}
	}
	glEnd();
	glPopAttrib();
#endif
#endif
#endif
}

void ObjInst::renderEdges(const vec3f &offset, float pointSource, float dist) const {
	if (!model) return;
	if (!triextra) return;

	const std::vector<tri_t> &tri = model->getTris();
	const std::vector<vec3f> &vertex = model->getVertexes();
	const std::vector<edge_t> &edge = model->getEdges();

	//the real thing
	glBegin(GL_QUADS);
	for (int ei = 0; ei < (int)edge.size(); ei++) {
		const edge_t &e = edge[ei];

		//if the edges have matching visibility info then skip out
		if (triextra[e.t[0]].vis == triextra[e.t[1]].vis) {
			continue;
		}

		//get which of the two can be seen
		bool lit = triextra[e.t[0]].vis ? 0 : 1;

		//get our two tri info's
		//such that i = the lit face, j = the unlit face
		int i = e.t[lit];
		int j = e.t[!lit];

		//get our vertex info
		//k is the vertex subindex of the lit face
		//l = the vertex subindex of the unlit face
		int k = e.v[lit];
		int kk = (k + 1) % 3;

		const vec3f &v1 = vertex[tri[i].p[kk].v];
		const vec3f &v2 = vertex[tri[i].p[k].v];
		vec3f d1 = ((v1 * pointSource) - offset) * dist;
		vec3f d2 = ((v2 * pointSource) - offset) * dist;

		glVertex3fv(v1);
		glVertex3fv(v2);
		glVertex3fv(v2 + d2);
		glVertex3fv(v1 + d1);
	}
	glEnd();
}

void ObjInst::calcVisibleFaces(const vec4f &v) {
	if (!model) return;
	if (!triextra) return;
	
	const std::vector<tri_t> &tri = model->getTris();

	for (int i = 0; i < (int)tri.size(); i++) {
		//set the appropriate bit for this face
		//dot product the incoming point with the plane equation to test which side its on
		//this takes advantage of the v.w coefficient for infinite sources (vectors)
		triextra[i].vis = dot(tri[i].plane, v) > -0.00001f;
	}
}

bool ObjInst::setModel(ObjFile *model) {

	//unload our old model instance stuff
	//don't call unload() because it subsequently calls reset() which will lose our basis info
	if (triextra) {
		delete triextra;
		triextra = NULL;
	}

	//set the new model
	this->model = model;

	//if we're null then dont calc visibility
	if (!model) return true;

	strncpy(title, model->getTitle(), sizeof(title));

	//load up our new model instance stuff
	int visCount = (int)model->getTris().size();

	triextra = new triextra_t[visCount];
	if (!triextra) {
		model = NULL;
		return false;
	}
	for (triextra_t *t = triextra; t < triextra + visCount; t++) {
		t->vis = false;
		t->next = -1;
	}

	//initial bbox
	bbox = model->getBBox();

	return true;
}

//stripped down version of what is below, assuming a point light (w=1)
vec3f ObjInst::calcRelative(const vec3f &origin) {
	vec3f relativeOrigin = origin - basis.t;
	relativeOrigin = quatRotate(quatConj(basis.r), relativeOrigin);
	relativeOrigin.x /= scale.x;
	relativeOrigin.y /= scale.y;
	relativeOrigin.z /= scale.z;
	return relativeOrigin;
}

/**
 * [RS T][P]   [RSP + T]
 * [ 0 1][1] = [   1   ]
 *
 * [RS T][V]   [RSP]
 * [ 0 1][0] = [ 0 ]
 *
 * [RS T][v]   [RSv + Tw]
 * [ 0 1][w] = [   w    ]
 */
vec4f ObjInst::calcRelative(const vec4f &origin) {
	quatf relativeOrigin((vec3f&)origin - basis.t * origin.w, origin.w);
	(vec3f&)relativeOrigin = quatRotate(quatConj(basis.r), (vec3f&)relativeOrigin);
	relativeOrigin.x /= scale.x;
	relativeOrigin.y /= scale.y;
	relativeOrigin.z /= scale.z;
	return relativeOrigin;
}

void ObjInst::calcVtxs() {

	if (!model) return;

	//buffer all our vertex data:
	if (vtxextra) {
		delete vtxextra;
		vtxextra = NULL;
	}

	vtxextra = new vec3f[model->getVertexes().size()];
	if (!vtxextra) return;

	const std::vector<vec3f> &mdlVertex = model->getVertexes();

	for (int i = 0; i < (int)model->getVertexes().size(); i++) {
		vtxextra[i] = mdlVertex[i];
		vtxextra[i].x *= scale.x;
		vtxextra[i].y *= scale.y;
		vtxextra[i].z *= scale.z;
		vtxextra[i] = quatRotate(basis.r, vtxextra[i]);
		vtxextra[i] += basis.t;
	}

}

void ObjInst::updateBBox() {
	if (!model) return;

	const std::vector<vec3f> &mdlVertex = model->getVertexes();
	for (int i = 0; i < (int)mdlVertex.size(); i++) {
		if (!i) {
			bbox.min = bbox.max = vtxextra[i];
		} else {
			bbox.stretch(vtxextra[i]);
		}
	}
}

void ObjInst::buildBSPNode(int node, box3f box) {
	vec3f size = box.max - box.min;

	assert(node >= 0 && node < bspnodeCount);

	bspnode[node].tri = -1;
	bspnode[node].axis = -1;
	bspnode[node].mid = 0;

	int c0 = (node << 1) + 1;
	int c1 = (node << 1) + 2;
	if (c1 >= bspnodeCount) return;	//if c1 passes then all children pass

	//get the max dim . . 
	if (size.x < 0) size.x = -size.x;
	if (size.y < 0) size.y = -size.y;
	if (size.z < 0) size.z = -size.z;

	//figure out our axis of division
	int axis = 0;
	if (size.x > size.y) {
		if (size.x > size.z) {
			axis = 0;	//x
		} else {
			axis = 2;	//z
		}
	} else if (size.y > size.z) {
		axis = 1;	//y
	} else {
		axis = 2;	//z
	}
	bspnode[node].axis = axis;

	float mid = (box.max[axis] + box.min[axis]) * (0.4f + frand() * 0.2f);
	bspnode[node].mid = mid;

	box3f childbox;

	//lower
	childbox = box;
	childbox.max[axis] = mid;
	buildBSPNode(c0, childbox);

	//upper
	childbox = box;
	childbox.min[axis] = mid;
	buildBSPNode(c1, childbox);
}

void ObjInst::addTriToBSP(int i, const box3f &box, int node) {

	assert(node >= 0 && node < bspnodeCount);

	//TODO - assert that this tri doesnt belong to any face

	if (bspnode[node].axis != -1) {
		//msvc is complaining about using max[] and min[], so use ()'s instead
		if (box.max.operator()(bspnode[node].axis) < bspnode[node].mid) {
			//add to c0
			addTriToBSP(i, box, (node<<1)+1);
			return;
		} else if (box.min.operator()(bspnode[node].axis) > bspnode[node].mid) {
			//add to c1
			addTriToBSP(i, box, (node<<1)+2);
			return;
		}
	}

	int counter = 0;
	//add it to the end of this node's chain
	triextra[i].next = -1;
	if (bspnode[node].tri == -1) {
		//first on the list
		bspnode[node].tri = i;
	} else {
		//add to the end:
		int next = bspnode[node].tri;
		while (triextra[next].next != -1) {
			next = triextra[next].next;
			counter++;
		}
		triextra[next].next = i;
	}
}

bool ObjInst::buildBSPTree(int nodeCount) {
	if (bspnode) {
		delete bspnode;
		nodeCount = 0;
	}

	bspnode = new bspnode_t[nodeCount];
	if (!bspnode) return false;
	bspnodeCount = nodeCount;
	if (!bspnodeCount) return true;

	//first - add all the nodes into the first tree
	//next - recursively drop down?
	//no - first - make all leaves empty
	//next - cycle thru all tris and add them

	buildBSPNode(0, bbox);

	const std::vector<tri_t> &mdlTri = model->getTris();
	const std::vector<vec3f> &mdlVertex = model->getVertexes();

	for (int i = 0; i < (int)mdlTri.size(); i++) {
		box3f box( vtxextra[mdlTri[i].p[0].v] );
		box.stretch( vtxextra[mdlTri[i].p[1].v] );
		box.stretch( vtxextra[mdlTri[i].p[2].v] );
		addTriToBSP(i, box);
	}

	return true;
}


bool ObjInst::testSphereCollision(const vec3f &src, float rad, col_t *col) {
	box3f colvol(src);
	colvol.stretch(col->x);
	colvol.min -= vec3f(1,1,1);
	colvol.max += vec3f(1,1,1);

	if (!colvol.touches(bbox)) return false;

	return testTreeSphereCollision(src, rad, col, 0);
}

bool ObjInst::testTreeSphereCollision(const vec3f &src, float rad, col_t *col, int node) {
	assert (bspnode);	//no bsp node data!
	assert(node >= 0 && node < bspnodeCount);

	//see if we can test the children:

	bool collided = false;

//	if (!bbox.intersects(colvol)) continue;

	//for all triangles within the object
	const std::vector<tri_t> &mdlTri = model->getTris();
	const std::vector<vec3f> &mdlVertex = model->getVertexes();

	int testCount = 0;

	int next = bspnode[node].tri;
	while (next != -1) {

		const vec3f *pt[3];
		pt[0] = &vtxextra[ mdlTri[next].p[0].v ];
		pt[1] = &vtxextra[ mdlTri[next].p[1].v ];
		pt[2] = &vtxextra[ mdlTri[next].p[2].v ];

//			printf("testing tri %f %f %f / %f %f %f / %f %f %f\n",
//				VEC3ELEM(pt[0]), VEC3ELEM(pt[1]), VEC3ELEM(pt[2]));

		//test for a collision with that triangle
		float bcc[3];
		float frac = testSphereTriCollision(src, col->x, rad, pt, &col->p, bcc);

		if (frac < 1.f) {
			collided = true;

			col->vtx[0] = *pt[0];
			col->vtx[1] = *pt[1];
			col->vtx[2] = *pt[2];
			
			col->bcc[0] = bcc[0];
			col->bcc[1] = bcc[1];
			col->bcc[2] = bcc[2];

			col->restitution = restitution;
			col->friction = friction;

			if (frac == 0.f) {
				col->x = src;
				col->frac = 0;
				return true;			//we've stopped - no more collisions could occur =)
			}
			col->x = vec3lerp(src, col->x, frac);
			col->frac *= frac;
		}

		next = triextra[next].next;
	};

	if (bspnode[node].axis != -1) {
		if (src(bspnode[node].axis) < bspnode[node].mid ||
			col->x(bspnode[node].axis) < bspnode[node].mid)
		{
			collided |= testTreeSphereCollision(src, rad, col, (node<<1)+1);
		}

		if (src(bspnode[node].axis) > bspnode[node].mid ||
			col->x(bspnode[node].axis) > bspnode[node].mid)
		{
			collided |= testTreeSphereCollision(src, rad, col, (node<<1)+2);
		}
	}

	return collided;
}