#include <stdio.h>
#include <assert.h>

#include "lib/springmodel.h"
#include "lib/objinst.h"

void SpringModel::clear() {
	SpringObj::clear();
	model = NULL;
	tri = NULL;
}

void SpringModel::shutdown() {
	if (tri) free(tri);
	SpringObj::shutdown();	//looks like SpringObj::shutdown()'s call to clear() won't hit up the right virtual function...
	clear();
}

//todo - ensure GL_NORMALIZE is enabled
//and ensure your color is white
void SpringModel::render() {
	if (!model) return;

	bool useTC = model->getFlags() & bitflag(ObjFile::USE_TEXCOORDS);

	const std::vector<tri_t> &mdlTri = model->getTris();
	const std::vector<vec2f> &mdlTexCoord = model->getTexCoords();

	glBegin(GL_TRIANGLES);
	for (int i = 0; i < (int)mdlTri.size(); i++) {
		for (int j = 0; j < 3; j++) {
			if (useTC) {
				glTexCoord2fv(mdlTexCoord[mdlTri[i].p[j].t]);
			}

			//NOTICE - if we're gonna use normals then we must accum them per vertex
			//or realign them per face

//			printf("f[%d] v[%d]: v[%d]: p[%d]: %f %f %f\n",
//				i,j,mdlTri[i].p[j].v,vertex[mdlTri[i].p[j].v],
//				VEC3ELEM(parSys.particle[vertex[mdlTri[i].p[j].v]].x));

			glVertex3fv(vertex[mdlTri[i].p[j].v]->x);
		}
	}
	glEnd();
}

void SpringModel::renderEdges(const vec3f &offset, float pointSource, float dist) const {
	if (!model) return;

	const std::vector<tri_t> &mdlTri = model->getTris();
	const std::vector<vec3f> &mdlVertex = model->getVertexes();
	const std::vector<edge_t> &mdlEdge = model->getEdges();

	//the real thing
	glBegin(GL_QUADS);
	for (int ei = 0; ei < (int)mdlEdge.size(); ei++) {
		const edge_t &e = mdlEdge[ei];

		//if the edges have matching visibility info then skip out
		if (tri[e.t[0]].vis == tri[e.t[1]].vis) {
			continue;
		}

		//get which of the two can be seen
		bool lit = tri[e.t[0]].vis ? 0 : 1;

		//get our two mdlTri info's
		//such that i = the lit face, j = the unlit face
		int i = e.t[lit];
		int j = e.t[!lit];

		//get our mdlVertex info
		//k is the mdlVertex subindex of the lit face
		//l = the mdlVertex subindex of the unlit face
		int k = e.v[lit];
		int kk = (k + 1) % 3;

		const vec3f &v1 = vertex[mdlTri[i].p[kk].v]->x;
		const vec3f &v2 = vertex[mdlTri[i].p[k ].v]->x;
		vec3f d1 = ((v1 * pointSource) - offset) * dist;
		vec3f d2 = ((v2 * pointSource) - offset) * dist;

		glVertex3fv(v1);
		glVertex3fv(v2);
		glVertex3fv(v2 + d2);
		glVertex3fv(v1 + d1);
	}
	glEnd();
}

void SpringModel::calcVisibleFaces(const vec4f &v) {
	if (!model) return;
	
	const std::vector<tri_t> &mdlTri = model->getTris();

	for (int i = 0; i < (int)mdlTri.size(); i++) {
		//set the appropriate bit for this face
		//dot product the incoming point with the plane equation to test which side its on
		//this takes advantage of the v.w coefficient for infinite sources (vectors)
		tri[i].vis = dot(mdlTri[i].plane, v) > -0.00001f;
	}
}

bool SpringModel::setModel(ObjFile *model) {
	//unload our old model instance stuff
	if (!init(parSys, springSys)) return false;

	//set the new model
	this->model = model;

	//if we're null then dont calc visibility
	if (!model) return true;

	//load up our new model instance stuff

	//edge data:
	int triSize = (int)(sizeof(SMTri) * model->getTris().size());
	tri = (SMTri *)malloc(triSize);
	if (!tri) return false;
	memset(tri, 0, triSize);

	//vertex data:
	int vertexCount = (int)model->getVertexes().size();
	vertex.resize(vertexCount);

	//allocate vertices:
	int i;
	for (i = 0; i < vertexCount; i++) {
		vertex[i] = parSys->createParticle();
		if (!vertex[i]) return false;

		Particle *p = vertex[i];
		p->massInv = 1;
		p->x.set(0,0,0);
		p->v.set(0,0,0);
		p->clearForce();
	}

//set SPRINGTYPE to this for all points of the model to be strung with springs
#define	SPRING_ALL_POINTS	1

//set SPRINGTYPE to this for only the edges of the model to be strung with springs
#define	SPRING_EDGES		2

//set SPRINGTYPE to this for no springs at all
#define	SPRING_NONE			3

#define	SPRINGTYPE			SPRING_EDGES

	//spring data:
#if SPRINGTYPE == SPRING_ALL_POINTS
	int springCount = vertexCount;
#elif SPRINGTYPE == SPRING_EDGES
	int springCount = (int)model->getEdges().size();
#elif SPRINGTYPE == SPRING_NONE
	int springCount = 0;
#endif
	spring.resize(springCount);

	//allocate springs:
#if SPRINGTYPE == SPRING_ALL_POINTS

	i = 0;
	for (int v0 = 0; v0 < vertexCount-1; v0++) {
		for (int v1 = v0+1; v1 < vertexCount; v1++, i++) {
			
#elif SPRINGTYPE == SPRING_EDGES
	for (i = 0; i < springCount; i++) {
		const edge_t &srcEdge = model->getEdges()[i];
		int v0 = model->getTris()[srcEdge.t[0]].p[srcEdge.v[0]].v;
		int v1 = model->getTris()[srcEdge.t[1]].p[srcEdge.v[1]].v;

#elif SPRINGTYPE == SPRING_NONE

	if (0) {
		int v0 = 0, v1 = 0;

#endif
		spring[i] = springSys->createSpring();
		if (spring[i] == -1) {
			printf("failed to create a spring\n");
			return false;
		}

		//match up the vertexes
		//make sure the vertices have been already allocated
		Spring *s = &springSys->spring[spring[i]];
		s->par[0] = vertex[v0];
		s->par[1] = vertex[v1];
		s->restlen = 0;
		s->ks = 700.f;
		s->kd = 10.f;
#if SPRINGTYPE == SPRING_ALL_POINTS
	}
#endif
	}

	//reset the model state - reposition all vertices and re-stretch all springs
		//first reset vertices so that they'll be in the right position for the edges
	resetModelVertexes();
		//next reset the edges - this should only need to ever be done here
		//unless you want to deform an object and make it stay in that deformed shape =)
	resetSpringLengths();

	return true;
}

void SpringModel::resetModelVertexes() {
	assert(model);

	const std::vector<vec3f> &mdlVertex = model->getVertexes();

	//align all vertices with their parent model vertex positions
	for (unsigned int i = 0; i < mdlVertex.size(); i++) {
		vertex[i]->x = mdlVertex[i];
	}
}

void SpringModel::resetSpringLengths() {
	assert(model);
	
	SpringObj::resetSpringLengths();

	const std::vector<edge_t> &mdlEdge = model->getEdges();
	const std::vector<tri_t> &mdlTri = model->getTris();

#if 0	//handled in super::
/*
	//allocate springs:
#ifdef SPRING_ALL_POINTS
	for (int v0 = 0; v0 < vertexCount-1; v0++) {
		for (int v1 = v0+1; v1 < vertexCount; v1++) {
#else
	for (i = 0; i < edgeCount; i++) {
		const edge_t &srcEdge = model->getEdges()[i];
		int v0 = model->getTris()[srcEdge.t[0]].p[srcEdge.v[0]].v;
		int v1 = model->getTris()[srcEdge.t[1]].p[srcEdge.v[1]].v;

		spring[i] = springSys.createSpring();
		if (spring[i] == -1) {
			printf("failed to create a spring\n");
			return false;
		}
#endif
*/
	//compute rest lengths of all spring edges
#if SPRINGTYPE == SPRING_ALL_POINTS

	int vertexCount = model->getVertexes().size();
	int i = 0;
	for (int v0 = 0; v0 < vertexCount-1; v0++) {
		for (int v1 = v0+1; v1 < vertexCount; v1++, i++) {
		Spring *s = &springSys->spring[spring[i]];
		assert(s->active);

#elif SPRINGTYPE == SPRING_EDGES

	for (int i = 0; i < (int)mdlEdge.size(); i++) {
		Spring *s = &springSys->spring[spring[i]];
		assert(s->active);
		int v0 = mdlTri[mdlEdge[i].t[0]].p[mdlEdge[i].v[0]].v;
		int v1 = mdlTri[mdlEdge[i].t[1]].p[mdlEdge[i].v[1]].v;

#elif SPRINGTYPE == SPRING_NONE

	int i;
	if (0) {
		int v0 = 0, v1 = 0;
		Spring *s = NULL;

#endif

		Particle *p0 = vertex[v0];
		Particle *p1 = vertex[v1];
		assert(p0 && p1);
		s->restlen = (p0->x - p1->x).len();
//		printf("spring %d restlen: %f\n", edge[i], s->restlen);
#if SPRINGTYPE == SPRING_ALL_POINTS
	}
#endif
	}
#endif

	for (int i = 0; i < (int)model->getTris().size(); i++) {
		int v[3];
		Particle *p[3];
		int j;
		for (j = 0; j < 3; j++) {
			v[j] = mdlTri[i].p[j].v;
			p[j] = vertex[v[j]];
		}
		for (j = 0; j < 3; j++) {
#ifdef TRIANGLE_CONSTRAINT
			int v0 = j;
			int v1 = (j + 1) % 3;
			int v2 = (j + 2) % 3;
			vec3f d01 = p[v1]->x - p[v0]->x;
			vec3f d12 = p[v2]->x - p[v1]->x;
			tri[i].angle[j] = d01 % d01;
			tri[i].lensq[j] = d01 % d12;
#else
			int a = (j + 2) % 3;
			int b = j;
			int c = (j + 1) % 3;
			vec3f d1 = p[a]->x - p[b]->x;
			vec3f d2 = p[c]->x - p[b]->x;
			tri[i].angle[j] = dot(d1, d2) / (float)(d1.len() * d2.len());
#endif
		}
	}

	//and we should be good to go!
}

//physics

void SpringModel::applyForces() {

	const std::vector<tri_t> &mdlTri = model->getTris();
	for (int i = 0; i < (int)mdlTri.size(); i++) {
		int v[3];
		Particle *p[3];
		int j;
		for (j = 0; j < 3; j++) {
			v[j] = mdlTri[i].p[j].v;
			p[j] = vertex[v[j]];
		}
		for (j = 0; j < 3; j++) {
#ifdef TRIANGLE_CONSTRAINT
			int v0 = j;
			int v1 = (j + 1) % 3;
			int v2 = (j + 2) % 3;
			vec3f d0 = p[v0]->x - p[v1]->x;
			vec3f d1 = p[v1]->x - p[v2]->x;
			vec3f d2 = p[v2]->x - p[v0]->x;

//			tri[i].angle[j] = d01 % d01;
//			tri[i].lensq[j] = d01 % d12;

			//2(v0-v1)((v0-v1).(v0-v1)-l0)
			//	+ 2(v0-v2)((v0-v2).(v0-v2)-l2)
			//	+ (v1-v2)((v1-v0).(v2-v1)-th0)
			//	+ (v2-v1)((v2-v1).(v0-v1)-th1)
			//	+ (v2+v1-2v0)((v0-v2).(v1-v0)-th2)
			//
			//2(v0-v1)((v0-v1).(v0-v1)-l0)
			//	+ 2(v0-v2)((v0-v2).(v0-v2)-l2)
			//	+ (v1-v2)((v1-v0).(v2-v1)-th0)
			//	+ (v2-v1)((v2-v1).(v0-v1)-th1)
			//	+ (v2+v1-2v0)((v0-v2).(v1-v0)-th2)
			//
			//2(v0-v1)((v0-v1).(v0-v1)-l0)
			//	+ 2(v0-v2)((v0-v2).(v0-v2)-l2)
			//	+ (v1-v2)((v1-v0).(v2-v1)-th0)
			//	+ (v2-v1)((v2-v1).(v0-v1)-th1)
			//	+ ((v2-v0)+(v1-v0))((v0-v2).(v1-v0)-th2)
			//
			//2(v0-v1)((v0-v1).(v0-v1)-l0)
			//	+ 2(v0-v2)((v0-v2).(v0-v2)-l2)
			//	+ (v1-v2)((v1-v0).(v2-v1)-th0)
			//	- (v1-v2)((v2-v1).(v0-v1)-th1)
			//	+ ((v2-v0)+(v1-v0))((v0-v2).(v1-v0)-th2)
			//
			//2(v0-v1)((v0-v1).(v0-v1)-l0)
			//	+ 2(v0-v2)((v0-v2).(v0-v2)-l2)
			//	+ (v1-v2)((v1-v0).(v2-v1)-th0)
			//	- (v1-v2)((v2-v1).(v0-v1)-th1)
			//	+ (v2-v0)((v0-v2).(v1-v0)-th2)
			//  + (v1-v0)((v0-v2).(v1-v0)-th2)
			//
			//2(v0-v1)((v0-v1).(v0-v1)-l0)
			//	+ 2(v0-v2)((v0-v2).(v0-v2)-l2)
			//	+ (v1-v2)((v2-v1).(v1-v0) - th0 - (v2-v1).(v0-v1) + th1)
			//	+ (v2-v0)((v0-v2).(v1-v0)-th2)
			//  + (v1-v0)((v0-v2).(v1-v0)-th2)
			//
			//(v0-v1)(2(v0-v1).(v0-v1) - 2l0 - (v0-v2).(v1-v0) + th2)
			//	+ 2(v0-v2)((v0-v2).(v0-v2)-l2)
			//	+ (v1-v2)(th1 - th0)
			//	+ (v2-v0)((v0-v2).(v1-v0)-th2)
			//
			//(v0-v1)(2(v0-v1).(v0-v1) - (v0-v2).(v1-v0) + th2 - 2l0)
			//	+ (v0-v2)(2(v0-v2).(v0-v2) - 2l2)
			//	+ (v1-v2)(th1 - th0)
			//	- (v0-v2)((v0-v2).(v1-v0) - th2)
			//
			//    (v0-v1)(2*(v0-v1).(v0-v1) - 2*l0 - (v2-v0).(v0-v1) + th2)
			//	+ (v1-v2)(th1 - th0)
			//	+ (v2-v0)((v2-v0).(v0-v1) - th2 - 2(v2-v0).(v2-v0) + 2*l2)
			vec3f dif = d0 * (2.f * ((d0 % d0) - tri[i].lensq[v0]) - ((d2 % d0) - tri[i].angle[v2]))
					  + d1 * (tri[i].angle[v1] - tri[i].angle[v0])
					  + d2 * (((d2 % d0) - tri[i].angle[v2]) - 2.f * ((d2 % d2) - tri[i].lensq[v2]));
			p[v0]->f -= dif;

#else
			int a = (j + 2) % 3;
			int b = j;
			int c = (j + 1) % 3;
			vec3f d1 = p[a]->x - p[b]->x;
			vec3f d2 = p[c]->x - p[b]->x;
			float l1 = (float)d1.len();
			float l2 = (float)d2.len();
			float il1 = 1.f / l1;
			float il2 = 1.f / l2;
			float d1d2 = dot(d1, d2);

			float scale = 2.f * il1 * il2;

			//C = [unit(p[a]-p[b]) dot unit(p[c]-p[b]) - restAngle]^2
			//dC/dp[a]:
			vec3f f = d2 * (2.f * d1d2 * il1 * il2)
					 - d1 * (d1d2 * d1d2 * il1 * il1 * il1 * il2)
					 - d2 * tri[i].angle[j]
					 + d1 * (tri[i].angle[j] * il1 * il1);

			//damp it down and step against the gradeitn
			//TODO - incorperate dC/dt as a velocity damping
			p[a]->addForce(f * -0.1f);
#endif
		}
	}
}
